import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Self, cast

import tomlkit
from lightman_ai.core.exceptions import ConfigNotFoundError, InvalidConfigError, PromptNotFoundError
from pydantic import BaseModel, ConfigDict, PositiveInt, ValidationError, field_validator
from pydantic_core.core_schema import FieldValidationInfo

PROMPTS_SECTION = "prompts"
logger = logging.getLogger("lightman")


def read_config_from_file(*, config_section: str, path: str) -> dict[str, Any]:
    fpath = Path(path)
    if not fpath.exists():
        raise ConfigNotFoundError(f"`{path}` not found!")

    content = fpath.read_text()
    parsed_content = tomlkit.parse(content)

    return cast("dict[str, Any]", parsed_content.get(config_section, {}))


@dataclass
class PromptConfig:
    prompts: dict[str, str]

    @classmethod
    def get_config_from_file(cls, path: str) -> Self:
        config = read_config_from_file(config_section=PROMPTS_SECTION, path=path)
        return cls(prompts=config)

    def get_prompt(self, prompt: str) -> str:
        if prompt not in self.prompts:
            raise PromptNotFoundError(f"prompt `{prompt}` not found in config file")
        return self.prompts[prompt]


class FileConfig(BaseModel):
    prompt: str | None = None
    agent: str | None = None
    score_threshold: int | None = None
    service_desk_project_key: str | None = None
    service_desk_request_id_type: str | None = None
    model: str | None = None

    model_config = ConfigDict(extra="forbid")

    @field_validator("service_desk_project_key", "service_desk_request_id_type", mode="before")
    @classmethod
    def _cast_service_fields(cls, v: Any, info: FieldValidationInfo) -> Any:
        if v is None:
            return v
        if isinstance(v, int):
            return str(v)
        if isinstance(v, str):
            if v == "":
                return v
            try:
                int(v)
            except ValueError as err:
                raise ValueError(f"{info.field_name} must be a number!") from err
        return v

    @classmethod
    def get_config_from_file(cls, config_section: str, path: str) -> Self:
        try:
            config = read_config_from_file(config_section=config_section, path=path)
        except ConfigNotFoundError:
            logger.warning("Config file `%s` not found! Proceeding with empty config.", path)
            config = {}
        return cls(**config)


class FinalConfig(BaseModel):
    prompt: str
    agent: str
    score_threshold: PositiveInt
    model: str | None = None

    @classmethod
    def init_from_dict(cls, data: dict[str, Any]) -> Self:
        try:
            return cls(**data)
        except ValidationError as error:
            error_list = []
            for err in error.errors():
                error_list.append(f"`{err['loc'][0]}`: {err['msg']}")
            err_msg = f"Invalid configuration provided: [{','.join(error_list)}]"
            raise InvalidConfigError(err_msg) from error



from typing import override
from xml.etree import ElementTree

import httpx
import stamina
from httpx import Client
from lightman_ai.article.models import Article, ArticlesList
from lightman_ai.sources.base import BaseSource

_RETRY_ON = httpx.TransportError
_ATTEMPTS = 5
_TIMEOUT = 5


THN_URL = "https://feeds.feedburner.com/TheHackersNews"


class TheHackerNewsSource(BaseSource):
    @override
    def get_articles(self) -> ArticlesList:
        """Return the articles that are present in THN feed."""
        feed = self.get_feed()
        articles = self._xml_to_list_of_articles(feed)
        return ArticlesList(articles=articles)

    def get_feed(self) -> str:
        """Retrieve the TheHackerNews' RSS Feed."""
        for attempt in stamina.retry_context(
            on=_RETRY_ON,
            attempts=_ATTEMPTS,
            timeout=_TIMEOUT,
        ):
            with Client() as http_client, attempt:
                hacker_news_feed = http_client.get(THN_URL)
                hacker_news_feed.raise_for_status()
        return hacker_news_feed.text

    def _xml_to_list_of_articles(self, xml: str) -> list[Article]:
        root = ElementTree.fromstring(xml)
        channel = root.find("channel")
        assert channel
        items = channel.findall("item")

        parsed = []

        for item in items:
            title = item.findtext("title", default="").strip()
            description = self._clean(item.findtext("description", default="").strip())
            link = item.findtext("link", default="").strip()

            parsed.append(Article(title=title, description=description, link=link))
        return parsed

    @staticmethod
    def _clean(text: str) -> str:
        """Remove non-useful characters."""
        return text.replace("\\n", "").replace("       ", "")



from typing import override

from lightman_ai.ai.base.agent import BaseAgent
from lightman_ai.ai.gemini.exceptions import map_gemini_exceptions
from lightman_ai.article.models import SelectedArticlesList
from pydantic_ai.models.google import GoogleModel


class GeminiAgent(BaseAgent):
    """Class that provides an interface to operate with the Gemini model."""

    _class = GoogleModel
    _default_model_name = "gemini-2.5-pro-preview-05-06"

    @override
    def _run_prompt(self, prompt: str) -> SelectedArticlesList:
        with map_gemini_exceptions():
            result = self.agent.run_sync(prompt)
        return result.output



from typing import Any

from pydantic_settings import BaseSettings


class Settings(BaseSettings):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    OPENAI_ENCODING: str = "cl100k_base"
    PARALLEL_WORKERS: int = 5


settings = Settings()



import logging
from abc import ABC, abstractmethod
from typing import Never

from lightman_ai.article.models import SelectedArticlesList
from pydantic_ai import Agent
from pydantic_ai.models.google import GoogleModel
from pydantic_ai.models.openai import OpenAIModel


class BaseAgent(ABC):
    _class: type[OpenAIModel] | type[GoogleModel]
    _default_model_name: str

    def __init__(self, system_prompt: str, model: str | None = None, logger: logging.Logger | None = None) -> None:
        agent_model = self._class(model or self._default_model_name)
        self.agent: Agent[Never, SelectedArticlesList] = Agent(
            model=agent_model, output_type=SelectedArticlesList, system_prompt=system_prompt
        )
        self.logger = logger or logging.getLogger("lightman")

    def get_prompt_result(self, prompt: str) -> SelectedArticlesList:
        return self._run_prompt(prompt)

    @abstractmethod
    def _run_prompt(self, prompt: str) -> SelectedArticlesList: ...



from lightman_ai.ai.base.agent import BaseAgent
from lightman_ai.ai.gemini.agent import GeminiAgent
from lightman_ai.ai.openai.agent import OpenAIAgent

AGENT_MAPPING = {"openai": OpenAIAgent, "gemini": GeminiAgent}


def get_agent_class_from_agent_name(agent: str) -> type[BaseAgent]:
    if agent not in AGENT_MAPPING:
        raise ValueError(f"Agent '{agent}' is not recognized. Available agents: {list(AGENT_MAPPING.keys())}")
    return AGENT_MAPPING[agent]


AGENT_CHOICES = list(AGENT_MAPPING.keys())



import asyncio
import logging

from lightman_ai.ai.base.agent import BaseAgent
from lightman_ai.ai.utils import get_agent_class_from_agent_name
from lightman_ai.article.models import ArticlesList, SelectedArticle, SelectedArticlesList
from lightman_ai.integrations.service_desk.integration import (
    ServiceDeskIntegration,
)
from lightman_ai.sources.the_hacker_news import TheHackerNewsSource

logger = logging.getLogger("lightman")


def _get_articles() -> ArticlesList:
    return TheHackerNewsSource().get_articles()


def _classify_articles(articles: ArticlesList, agent: BaseAgent) -> SelectedArticlesList:
    return agent.get_prompt_result(prompt=str(articles))


def _create_service_desk_issues(
    selected_articles: list[SelectedArticle],
    service_desk_client: ServiceDeskIntegration,
    project_key: str,
    request_id_type: str,
) -> None:
    async def schedule_task(article: SelectedArticle) -> None:
        try:
            description = f"*Why is relevant:*\n{article.why_is_relevant}\n\n*Source:* {article.link}\n\n*Score:* {article.relevance_score}/10"
            await service_desk_client.create_request_of_type(
                project_key=project_key,
                summary=article.title,
                description=description,
                request_id_type=request_id_type,
            )
            logger.info("Created issue for article %s", article.link)
        except Exception:
            logger.exception("Could not create ServiceDesk issue: %s, %s", article.title, article.link)
            raise

    async def create_all() -> None:
        tasks = []
        for article in selected_articles:
            tasks.append(schedule_task(article))
        results = await asyncio.gather(*tasks, return_exceptions=True)
        errors = [result for result in results if isinstance(result, Exception)]
        if errors:
            raise ExceptionGroup("Could not create all ServiceDesk issues", errors)

    asyncio.run(create_all())


def lightman(
    agent: str,
    prompt: str,
    score_threshold: int,
    project_key: str | None = None,
    request_id_type: str | None = None,
    dry_run: bool = False,
    model: str | None = None,
) -> list[SelectedArticle]:
    articles: ArticlesList = _get_articles()

    agent_class = get_agent_class_from_agent_name(agent)
    agent_instance = agent_class(prompt, model, logger=logger)
    logger.info("Selected %s.", agent_instance)

    classified_articles = _classify_articles(
        articles=articles,
        agent=agent_instance,
    )

    selected_articles: list[SelectedArticle] = classified_articles.get_articles_with_score_gte_threshold(
        score_threshold
    )
    if selected_articles:
        logger.info("Found these articles: %s", selected_articles)
    else:
        logger.info("No articles found to be relevant. Total returned articles by AI %s", len(classified_articles))

    if not dry_run:
        if not project_key or not request_id_type:
            raise ValueError("Missing Service Desk's project key or request id type")

        service_desk_client = ServiceDeskIntegration.from_env()
        _create_service_desk_issues(
            selected_articles=selected_articles,
            service_desk_client=service_desk_client,
            project_key=project_key,
            request_id_type=request_id_type,
        )

    return selected_articles



import logging
from importlib import metadata

import click
from dotenv import load_dotenv
from lightman_ai.ai.utils import AGENT_CHOICES
from lightman_ai.constants import DEFAULT_CONFIG_FILE, DEFAULT_CONFIG_SECTION, DEFAULT_ENV_FILE
from lightman_ai.core.config import FileConfig, FinalConfig, PromptConfig
from lightman_ai.core.exceptions import ConfigNotFoundError, InvalidConfigError, PromptNotFoundError
from lightman_ai.core.sentry import configure_sentry
from lightman_ai.main import lightman

logger = logging.getLogger("lightman")


def get_version() -> str:
    """Read version from VERSION file."""
    return metadata.version("lightman-ai")


@click.group()
@click.version_option(version=get_version(), prog_name="lightman-ai")
def entry_point() -> None:
    pass


@entry_point.command()
@click.option("--agent", type=click.Choice(AGENT_CHOICES), help=("Which agent to use"))
@click.option(
    "--prompt-file",
    type=str,
    default=DEFAULT_CONFIG_FILE,
    help=(f"Location of the config file containing the prompts. Defaults to `{DEFAULT_CONFIG_FILE}`."),
)
@click.option("--prompt", type=str, help=("Which prompt to use"))
@click.option("--model", type=str, default=None, help=("Which model to use. Must be set in conjunction with --agent."))
@click.option(
    "--score",
    type=int,
    help=("The minimum score relevance that an article needs to have to be considered relevant"),
    default=None,
)
@click.option(
    "--config-file",
    type=str,
    default=DEFAULT_CONFIG_FILE,
    help=(f"The config file path. Defaults to `{DEFAULT_CONFIG_FILE}`."),
)
@click.option(
    "--config",
    type=str,
    default=DEFAULT_CONFIG_SECTION,
    help=(f"The config settings to use. Defaults to `{DEFAULT_CONFIG_SECTION}`."),
)
@click.option(
    "--env-file",
    type=str,
    default=DEFAULT_ENV_FILE,
    help=(f"Path to the environment file. Defaults to `{DEFAULT_ENV_FILE}`."),
)
@click.option(
    "--dry-run",
    is_flag=True,
    help=(
        "When set, runs the script without publishing the results to the integrated services, just shows them in stdout."
    ),
)
def run(
    agent: str,
    prompt: str,
    prompt_file: str,
    model: str | None,
    score: int | None,
    config_file: str,
    config: str,
    env_file: str,
    dry_run: bool,
) -> int:
    """
    Entrypoint of the application.

    Holds no logic. It calls the main method and returns 0 when succesful .
    """
    load_dotenv(env_file)
    configure_sentry()
    try:
        prompt_config = PromptConfig.get_config_from_file(path=prompt_file)
        config_from_file = FileConfig.get_config_from_file(config_section=config, path=config_file)
        final_config = FinalConfig.init_from_dict(
            data={
                "agent": agent or config_from_file.agent,
                "prompt": prompt or config_from_file.prompt,
                "score_threshold": score or config_from_file.score_threshold,
                "model": model or config_from_file.model,
            }
        )

        prompt_text = prompt_config.get_prompt(final_config.prompt)
    except (InvalidConfigError, PromptNotFoundError, ConfigNotFoundError) as err:
        raise click.BadParameter(err.args[0]) from None

    relevant_articles = lightman(
        agent=final_config.agent,
        prompt=prompt_text,
        score_threshold=final_config.score_threshold,
        dry_run=dry_run,
        project_key=config_from_file.service_desk_project_key,
        request_id_type=config_from_file.service_desk_request_id_type,
        model=final_config.model,
    )
    relevant_articles_metadata = [f"{article.title} ({article.link})" for article in relevant_articles]
    logger.warning("Found these articles: \n- %s", "\n- ".join(relevant_articles_metadata))
    return 0



from abc import ABC
from typing import override

from pydantic import BaseModel


class BaseArticle(BaseModel, ABC):
    """Base abstract class for all Articles."""

    title: str
    link: str

    @override
    def __eq__(self, value: object) -> bool:
        if not isinstance(value, BaseArticle):
            return False

        return self.link == value.link

    @override
    def __hash__(self) -> int:
        return hash(self.link.encode())


class SelectedArticle(BaseArticle):
    why_is_relevant: str
    relevance_score: int


class Article(BaseArticle):
    title: str
    description: str


class BaseArticlesList[TArticle: BaseArticle](BaseModel):
    articles: list[TArticle]

    def __len__(self) -> int:
        return len(self.articles)

    @property
    def titles(self) -> list[str]:
        return [new.title for new in self.articles]

    @property
    def links(self) -> list[str]:
        return [new.link for new in self.articles]


class SelectedArticlesList(BaseArticlesList[SelectedArticle]):
    """
    Model that holds all the articles that were selected by the AI model.

    It saves the minimum information so that they are identifiable.
    """

    def get_articles_with_score_gte_threshold(self, score_threshold: int) -> list[SelectedArticle]:
        if not score_threshold > 0:
            raise ValueError("score threshold must be > 0.")
        return [article for article in self.articles if article.relevance_score >= score_threshold]


class ArticlesList(BaseArticlesList[Article]):
    """Model that saves articles with all their information."""



import logging
import os
from importlib import metadata

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration


def configure_sentry() -> None:
    """Configure Sentry for error tracking and performance monitoring using env vars with fallbacks."""
    try:
        if not os.getenv("SENTRY_DSN"):
            logging.getLogger("lightman").info("SENTRY_DSN not configured, skipping Sentry initialization")
            return

        # Logging level from ENV
        logging_level_str = os.getenv("LOGGING_LEVEL", "ERROR").upper()
        logging_level = getattr(logging, logging_level_str, logging.ERROR)

        # Set up logging integration
        sentry_logging = LoggingIntegration(level=logging.INFO, event_level=logging_level)

        sentry_sdk.init(
            release=metadata.version("lightman-ai"),
            integrations=[sentry_logging],
        )
    except Exception as e:
        logging.getLogger("lightman").warning("Could not instantiate Sentry! %s.\nContinuing with the execution.", e)
